""" Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """

import tornado.ioloop
import tornado.web
from tornado.httpserver import HTTPServer
import ssl
import sys
import os
import OpenSSL
from subprocess import check_call
import signal
import threading, time


server = None

pid = None


def update_proxy_details_in_thread():
    # Total 30 secs
    max_retry = 6
    # in secs
    sleep_in_retry = 5
    current_retry = 0
    from handlerworkers.configprops import get_config
    from utilities.util import update_env_with_proxy_details
    while current_retry < max_retry:
        try:
            print("Trying DAS proxy update", flush=True)
            update_env_with_proxy_details(get_config('PROXY', 'proxy_settings'))
            print("DAS proxy updated", flush=True)
            break
        except Exception as e:
            print(str(e))
            print("DAS proxy update thread sleeping for " + str(sleep_in_retry) + " secs", flush=True)
            time.sleep(sleep_in_retry)
            current_retry = current_retry + 1
    print("Exiting the DAS proxy update thread", flush=True)

class CSDASSRV:

    @staticmethod
    def shutdown():
        """
        The function is designed to kill this service, if it's running.  it does this by calling getpid()
        to get the current pid from the database. if a pid is returned, it's killed with a SIGTERM signal,
        terminating the service immediately.
        @return:
        """
        global server
        if server is not None:
            server.stop()
        tornado.ioloop.IOLoop.current().stop()
        try:
            if pid:
                os.kill(pid, signal.SIGTERM)
        except ProcessLookupError as error:
            pass
        check_call(['pkill', '-f', 'csdassrv.py'])
        exit(0)


    @staticmethod
    def start():
        """
        This function starts the csdassrv service.  In addtion to that, it also saves the
        active process id (pid) to the database, so the service can be shutdown programmatically.
        @return: None
        """
        from handlers import (
            apikeyshandler,
            appliancekeyshandler,
            usershandler,
            tokenhandler,
            ldaphandler,
            authenticationhandler,
            nfahandler,
            nfauserhandler,
            nfamethodhandler,
            confighandler,
            licensehandler,
            passwordhandler,
            clusterhandler,
            mqlistener,
            utilshandler,
            meahandler,
            radiushandler,
            upgradehandler
        )
        from utilities.propsmgr import PropertiesManager
        from utilities import cspropconstants as csp
        from handlerworkers.fdnclient import start_fdn_client

        mqlistener.start_listener()
        start_fdn_client()

        # Update env with proxy details in separate thread
        thread = threading.Thread(name='thread-das-proxy-update', target=update_proxy_details_in_thread)
        thread.start()

        app = tornado.web.Application([
            (r"/api-keys/?", apikeyshandler.APIKeysHandler),
            (r"/api-keys/validate/?", apikeyshandler.APIKeysValidateHandler),
            (r"/apikeys/?", appliancekeyshandler.ApplianceKeysHandler),
            (r"/users/?", usershandler.UsersHandler),
            (r"/users/(?P<uuid>[^\/]+)", usershandler.UsersHandler),
            (r"/query/users/?", usershandler.QueryUserHandler),
            (r"/authenticate/?", authenticationhandler.AuthenticateHandler),
            (r"/nfa-request", nfahandler.NFARequestHandler),
            (r"/nfa-reset", nfahandler.NFAResetHandler),
            (r"/nfauser/?", nfauserhandler.NFAUserHandler),
            (r"/nfamethod/?", nfamethodhandler.NFAMethodHandler),
            (r"/nfaconfig/?", nfamethodhandler.NFAConfigHandler),
            (r"/config/?", confighandler.ConfigHandler),
            (r"/token/?", tokenhandler.TokenHandler),
            (r"/ldap/?", ldaphandler.LDAPHandler),
            (r"/license/?", licensehandler.LicenseHandler),
            (r"/token/(?P<jti>[^\/]+)/?", tokenhandler.TokenHandler),
            (r"/password/?", passwordhandler.PasswordHandler),
            (r"/cluster/?", clusterhandler.ClusterHandler),
            (r"/cluster/health/?", clusterhandler.ClusterHealthHandler),
            (r"/clusteroperation/?", clusterhandler.HAHandler),
            (r"/upgrade", upgradehandler.UpgradeHandler),
            (r"/clusteroperation/(?P<id>[^\/]+)/?", clusterhandler.HAHandler),
            (r"/clustercommand/?", clusterhandler.ClusterCommandHandler),
            (r"/clustercommand/(?P<id>[^\/]+)/?", clusterhandler.ClusterCommandHandler),
            (r"/cluster/shareddata/?", clusterhandler.ClusterSharedDataHandler),
            (r"/utils/cloud/?", utilshandler.CloudHandler),
            (r"/utils/log-collect/?", utilshandler.LogCollectHandler),
            (r"/utils/log-download/?", utilshandler.LogDownloadHandler),
            (r"/mea/backup/?", meahandler.MEAHandler),
            (r"/mea/restore/?", meahandler.MEAHandler),
            (r"/execute/action/?", authenticationhandler.AuthenticatedActionHandler),
            (r"/radius/?", radiushandler.RadiusHandler),
            (r"/password-policy/?", passwordhandler.PasswordPolicyHandler)

        ], static_path=os.path.join(os.path.dirname(__file__), '..', "static"))

        sslctx = ssl.SSLContext(ssl.PROTOCOL_TLS)
        sslctx.options |= ssl.OP_NO_SSLv2
        sslctx.options |= ssl.OP_NO_SSLv3
        sslctx.options |= ssl.OP_NO_TLSv1
        sslctx.options |= ssl.OP_NO_TLSv1_1
        sslctx.options |= ssl.OP_NO_TLSv1_2
        sslctx.options |= OpenSSL.SSL.OP_CIPHER_SERVER_PREFERENCE

        cert = '{0}'.format(PropertiesManager().get(csp.P_CERTLOCATION))
        key = '{0}'.format(PropertiesManager().get(csp.P_KEYLOCATION))
        sslctx.load_cert_chain(cert, keyfile=key)

        global server
        # 200MB buffer size
        server = HTTPServer(app, ssl_options=sslctx, max_buffer_size=209715200)
        port = PropertiesManager().get(csp.P_PORT)
        server.listen(port)

        global pid
        pid = os.getpid()
        main_loop = tornado.ioloop.IOLoop.current()

        main_loop.start()


def main():
    """
    This is, obviously, the main() function for the CSDASSRV REST service.
    @return:
    """
    rootpath = os.path.join(os.path.dirname(os.path.abspath(__file__)), "../")
    os.chdir(rootpath)
    sys.path.append(os.path.join(os.path.dirname(__file__), "../"))

    # Include common packages in das service
    sys.path.append('/opt/cyops/python_packages/fortisoar')

    error_str = '**** Parameter Required - <start/stop> ****'
    if len(sys.argv) <= 1:
        quit()

    arg = str(sys.argv[1]).lower()
    if arg != 'start' and arg != 'stop':
        quit()

    svc = CSDASSRV()
    if arg == 'start':
        svc.start()
    else:
        svc.shutdown()

if __name__ == '__main__':
    main()
